//************************************************************************
// Example10Minimum.java
//
// We implement several algorithms for finding minima of functions of
// several variables. For arbitrary real functions we simply try many
// random points in a prescribed region and take the best of them. For
// continuous functions, we start from a given point and try to improve
// it more and more by changing the point at each stage in the most
// promising direction out of all directions of the coordinate axis.
// For differentiable functions we use the gradient to make the
// improvement.
//  
// Our main point here is to demonstrate the usage of abstract classes.
// Thus the classes of functions mentioned above are abstract because
// we lack a method of evaluating the function (and its partial derivatives
// when it is known to be differentiable) at a given point as long as it is
// not defined. Only when we are given a specific class of functions can
// we perform this calculation.
//************************************************************************

public class Example10Minimum
{
   public static void main (String[] args)
   {
      // c1 is -2 on the rectangle [0, 0.45] x [0, 0.55] and 0 elsewhere
      Characteristic c1 = new Characteristic (-2, new double[] {0, 0},
                                                  new double[] {0.45, 0.55});
      Characteristic c2 = new Characteristic (-3, new double[] {0.35, 0.45},
                                                  new double[] {1, 1});
      Characteristic c3 = new Characteristic (-4, new double[] {0.3, 0.5},
                                                  new double[] {0.4, 0.56});
      SimpleFunction simp = new SimpleFunction (new Characteristic[]
                                                  {c1, c2, c3});

      AbsLin al1 = new AbsLin (new double[] {3, 2}, -7);  // |3x + 2y - 7|
      AbsLin al2 = new AbsLin (new double[] {4, 7}, -18); // |4x + 7y - 18|
      AbsLin al3 = new AbsLin (new double[] {-4, 5}, -6); // |-4x + 5y - 6|
      SumAbsLin sumAbs = new SumAbsLin (new AbsLin[] {al1, al2, al3});

      Monomial mon1 = new Monomial (3, new int[] {4, 0});  // 3 x^4
      Monomial mon2 = new Monomial (1, new int[] {2, 2});  // x^2 y^2
      Monomial mon3 = new Monomial (2, new int[] {0, 2});  // 2 y^2
      Polynomial poly = new Polynomial (new Monomial[] {mon1, mon2, mon3});

      double[] coordsOfMyLower = {0, 0};
      double[] coordsOfMyUpper = {1, 1};
      Point myLower = new Point (new double[] {0, 0});
      Point myUpper = new Point (new double[] {1, 1});
      System.out.println (simp.min (myLower, myUpper, 10000) + "\n" +
                          sumAbs.min (myLower, myUpper, 10000) + "\n" +
                          sumAbs.localMin (myUpper, 1, 10000) + "\n" +
                          poly.min (myLower, myUpper, 10000) + "\n" +
                          poly.localMin (myLower, 1, 10000));
   }// method main
}// class Example10Minimum


class Point
{
   private double[] coordinates;

   Point (double[] coordinates)
   {
      this.coordinates = new double[coordinates.length];
      for (int j=0; j < coordinates.length; j++)
         this.coordinates[j] = coordinates[j];
   }// method Point

   Point (Point P)
   {this (P.coordinates);}

   double[] getCoordinates()
   {return coordinates;} 

   void setCoordinate (int j, double value)
   {coordinates [j] = value;} 

   int getDimension()
   {return coordinates.length;} 

   double norm() // returns the distance of a point from the origin
   {
      double sum = 0;
      for (int i=0; i < getDimension(); i++)
         sum += Math.pow (getCoordinates()[i], 2);
      return Math.sqrt (sum);
   }// method norm

   public String toString()
   {
      String result = "(";
      for (int j=0; j < coordinates.length - 1; j++)
         result += coordinates[j] + ", ";
      return result + coordinates[coordinates.length - 1] + ")";
   }// method toString
}// class Point


abstract class RealFunction
{
   RealFunction()
   {}

   abstract double valueAt(Point P);

   Point min (Point lower, Point upper, int iterations)
   {
      Point currentBest = new Point(lower);
      Point next = new Point (lower);
      double temp;
      for (int i=1; i <= iterations; i++)
      {
         for (int j=0; j < lower.getDimension(); j++)
         {
            temp = lower.getCoordinates()[j] +
                   (upper.getCoordinates()[j] - lower.getCoordinates()[j])
                   * Math.random(); 
            next.setCoordinate(j, temp);
         } 
         if (valueAt(next) < valueAt(currentBest))
            currentBest = new Point(next);
       }
       return currentBest;
   }// method min
}// class RealFunction


abstract class ContinuousFunction extends RealFunction
{
   ContinuousFunction()
   {}

   Point localMin (Point initial, double initialStepSize, int iterations)
   {
      Point currentBest = new Point(initial);
      for (int i=1; i <= iterations; i++)
      {
         double bestValue = valueAt (currentBest);
         int bestCoordinate = -1, upOrDown = -1;
         Point next = new Point (currentBest);
         for (int j=0; j < next.getDimension(); j++)
         {
            for (int k=0; k <= 1; k++)
            {
               next.setCoordinate (j, currentBest.getCoordinates()[j] +
                                     (2 * k - 1) * initialStepSize / Math.sqrt(i));
               if (valueAt(next) < bestValue)
               {
                  bestValue = valueAt(next);
                  bestCoordinate = j;
                  upOrDown = k;
               }
            }
            next.setCoordinate(j, currentBest.getCoordinates()[j]);
         }
         if (bestCoordinate >= 0)
            currentBest.setCoordinate (bestCoordinate,
                        currentBest.getCoordinates()[bestCoordinate] +
                        (2 * upOrDown - 1) * initialStepSize / Math.sqrt(i));
      }
      return currentBest;
   }// method localMin
}// class ContinuousFunction


abstract class DifferentiableFunction extends ContinuousFunction
{
   DifferentiableFunction()
   {}

   abstract double[] derivativesAt (Point P);

   Point localMin (Point initial, double initialStepSize, int iterations)
   {
      Point current = new Point (initial);
      for (int i=1; i <= iterations; i++)
      {
         double denom = (new Point (derivativesAt(current))).norm();
         denom = Math.sqrt(i) * Math.max (denom, 1E-6);
         for (int j=0; j < current.getDimension(); j++)
         {
            double temp = current.getCoordinates()[j]
                          - derivativesAt(current)[j]
                            * initialStepSize / denom;
            current.setCoordinate (j, temp);
         }
      }
      return current;
   }// method localMin
}// class DifferentiableFunction


class Characteristic extends RealFunction
{
   private double   value;
   private double[] lower, upper;

   Characteristic (double value, double[] lower, double[] upper)
   {
      super();
      this.value = value;
      this.lower = lower;
      this.upper = upper;
   }// method Monomial

   double valueAt (Point P)
   {
      boolean inside = true;
      for (int j=0; j < P.getDimension(); j++)
         inside &= P.getCoordinates()[j] >= lower[j] &&
                   P.getCoordinates()[j] <= upper[j];
      if (inside)
         return value;
      else
         return 0;
   }// method valueAt
}// class Characteristic


class SimpleFunction extends RealFunction
{
   private Characteristic[] terms;

   SimpleFunction (Characteristic[] terms)
   {
      super();
      this.terms = terms;
   }// method SimpleFunction

   double valueAt (Point P)
   {
      double result = 0;
      for (int j=0; j < terms.length; j++)
         result += terms[j].valueAt (P);
      return result;
   }// method valueAt
}// class SimpleFunction


class AbsLin extends ContinuousFunction
{
   private double[] coefficients;
   private double   free;

   AbsLin (double[] coefficients, double free)
   {
      super();
      this.coefficients = coefficients;
      this.free = free;
   }// method AbsLin

   double valueAt (Point P)
   {
      double result = 0;
      for (int j=0; j < coefficients.length; j++)
         result += coefficients[j] * P.getCoordinates()[j];
      return Math.abs (result + free);
   }// method valueAt
}// class AbsLin


class SumAbsLin extends ContinuousFunction
{
   private AbsLin[] terms;

   SumAbsLin (AbsLin[] terms)
   {
      super();
      this.terms = terms;
   }// method SumAbsLin

   double valueAt (Point P)
   {
      double result = 0;
      for (int j=0; j < terms.length; j++)
         result += terms[j].valueAt (P);
      return result;
   }// method valueAt
}// class SumAbsLin

class Monomial extends DifferentiableFunction
{
   private double coefficient;
   private int[] power;

   Monomial (double coefficient, int[] power)
   {
      super();
      this.coefficient = coefficient;
      this.power = new int[power.length];
      for (int j=0; j < power.length; j++)
         this.power[j] = power[j];
   }// method Monomial

   double valueAt (Point P)
   {
      double value = coefficient;
      for (int j=0; j < power.length; j++)
         value *= Math.pow (P.getCoordinates()[j], power[j]);
      return value;
   }// method valueAt

   double[] derivativesAt (Point P)
   {
      double[] derivative = new double [P.getDimension()];
      int[] temp = new int[power.length];
      for (int j=0; j < power.length; j++)
         temp[j] = power[j];
      for (int j=0; j < power.length; j++)
         if (power[j] > 0)
         {
            temp[j]--;
            derivative[j] = (new Monomial (coefficient * power[j], temp)).
                            valueAt(P);
            temp[j]++;
         }
         else
            derivative[j] = 0;
      return derivative;
   }// method derivativesAt
}// class Monomial


class Polynomial extends DifferentiableFunction
{
   private Monomial[] terms;

   Polynomial (Monomial[] terms)
   {
      super();
      this.terms = terms;
   }// method Polynomial

   double valueAt (Point P)
   {
      double result = 0;
      for (int j=0; j < terms.length; j++)
         result += terms[j].valueAt (P);
      return result;
   }// method valueAt

   double[] derivativesAt (Point P)
   {
      double[] result = new double[P.getDimension()];
      for (int j=0; j < terms.length; j++)
         for (int k=0; k < P.getDimension(); k++)
            result[k] += terms[j].derivativesAt(P)[k];
      return result;
   }// method derivativesAt
}// class Polynomial
